#!/usr/bin/env python
# -*- coding: utf-8 -*-
# cython: language_level=3, always_allow_keywords=True

"""
Purpose
=======

:program:`rul4` is a script that can be used to render an UL4 template.


The ``globals`` object
======================

Inside the template the object ``globals`` (an instance of the class
:class:`Globals`) will be available to make database connections, load and save
files, compile templates, access environment variables and parameters etc.
However access to those features can be switched off via command line
options.


Options
=======

:program:`rul4` supports the following options:

.. program:: rul4

.. option:: templates

	One or more template files. A file named ``-`` will be treated as
	standard input. The first file in the list is the main template, i.e. the
	one that gets rendered. All templates will be available in the main
	template as the ``globals.templates`` dictionary. The keys are the base names
	of the files (i.e. ``foo.ul4`` will be ``globals.templates.foo``; stdin will
	be ``globals.templates.stdin``).

.. option:: --oracle <flag>

	Provide the method :meth:`Globals.oracle` (as ``globals.oracle``) to the
	template? If switched off ``globals.oracle`` will be :const:`None`.

	(Allowed values are ``false``, ``no``, ``0``, ``true``, ``yes`` or ``1``;
	the default is ``true``)

.. option:: --sqlite <flag>

	Provide the method :meth:`Globals.sqlite` (as ``globals.sqlite``) to the
	template? If switched off ``globals.sqlite`` will be :const:`None`.

	(Allowed values are ``false``, ``no``, ``0``, ``true``, ``yes`` or ``1``;
	the default is ``true``)

.. option:: --mysql <flag>

	Provide the method :meth:`Globals.mysql` (as ``globals.mysql``) to the
	template? If switched off ``globals.mysql`` will be :const:`None`.

	(Allowed values are ``false``, ``no``, ``0``, ``true``, ``yes`` or ``1``;
	the default is ``true``)

.. option:: --redis <flag>

	Provide the method :meth:`Globals.redis` (as ``globals.redis``) to the
	template? If switched off ``globals.redis`` will be :const:`None`.

	(Allowed values are ``false``, ``no``, ``0``, ``true``, ``yes`` or ``1``;
	the default is ``true``)

.. option:: --system <flag>

	Provide the method :meth:`Globals.system` (as ``globals.system``) to the
	template? If switched off ``globals.system`` will be :const:`None`.

	(Allowed values are ``false``, ``no``, ``0``, ``true``, ``yes`` or ``1``;
	the default is ``true``)

.. option:: --load <flag>

	Provide the method :meth:`Globals.load` (as ``globals.load``) to the
	template? If switched off ``globals.load`` will be :const:`None`.

	(Allowed values are ``false``, ``no``, ``0``, ``true``, ``yes`` or ``1``;
	the default is ``true``)

.. option:: --save <flag>

	Provide the method :meth:`Globals.save` (as ``globals.save``) to the
	template? If switched off ``globals.save`` will be :const:`None`.

	(Allowed values are ``false``, ``no``, ``0``, ``true``, ``yes`` or ``1``;
	the default is ``true``)

.. option:: --compile <flag>

	Provide the method :meth:`Globals.compile` (as ``globals.compile``) to the
	template? If switched off ``globals.compile`` will be :const:`None`.

	(Allowed values are ``false``, ``no``, ``0``, ``true``, ``yes`` or ``1``;
	the default is ``true``)

.. option::  -e <encoding> , --encoding <encoding>

	The encoding of the templates files (default ``utf-8``)

.. option::  -w <value>, --whitespace <value>

	Specifies how to handle whitespace in the template (Allowed values are
	``keep``, ``strip``, or ``smart``). This can of course be overwritten with
	the template tag ``<?whitespace ...?>`` in the template files.

.. option:: -D, --define

	Defines an additional variable that will be available inside the template
	(e.g. the variable ``foo`` will be available as ``globals.vars.foo``).
	``-D`` can be specified multiple times. The following formats are supported:

	``var``
		Defines ``var`` as an empty string;

	``var=value``
		Defines ``var`` as the string ``value``;

	``var:type``
		Defines ``var`` as an empty variable of the type ``type``;

	``var:type=value``
		Defines ``var`` as a variable of the type ``type`` with the value
		``value``.

	``type`` can be any of the following:

	``int``
		``value`` is an integer value.

	``float``
		``value`` is a float value.

	``bool``
		``value`` is a boolean value. ``0``, ``no``, ``false``, ``False`` or
		the empty string will be recognized as false and ``1``, ``yes``,
		``true`` or ``True`` will be recognized as true.

	``str``
		``value`` is a string.

	``oracle``
		``value`` will be a connection to an Oracle database, e.g.::

			-Ddb:oracle=user/password@database

	``sqlite``
		``value`` is a connection to an SQLite database.

	``mysql``
		``value`` is a connection to a MySQL database.

	``redis``
		``value`` will be a connection to an Redis database, e.g.::

			-Ddb:redis=192.168.123.1:6379/42

		The port (i.e. the ``6379`` in the above value) is optional and
		defaults to 6379. The database number (i.e. the ``42`` in the above
		value) is also optional and defaults to 0.


Example
=======

This example shows how to connect to an Oracle database and output the content
of a ``person`` table into an XML file.

Suppose we have a database table that looks like this:

.. sourcecode:: sql

	create table person
	(
		id integer not null,
		firstname varchar2(200),
		lastname varchar2(200)
	);

Then we can use the following template to output the table into an XML file:

.. sourcecode:: xml

	<?xml version='1.0' encoding='utf-8'?>
	<?code db = globals.oracle("user/password@database')?>
	<persons>
		<?for p in db.query("select id, firstname, lastname from person order by 2, 1")?>
			<person id="<?printx p.id?>">
				<firstname><?printx p.firstname?></firstname>
				<lastname><?printx p.lastname?></lastname>
			</person>
		<?end for?>
	</persons>

If we put the template into the file :file:`person.ul4` we can call
:program:`rul4` like this:

.. sourcecode:: bash

	rul4 person.ul4 >person.xml

We could also pass the connection to our database via the :option:`-D` option
and disallow the script to make any database connections itself or execute any
system commands:

.. sourcecode:: bash

	rul4 person.ul4 -Ddb:oracle=user/password@database --oracle=0 --sqlite=0 --mysql=0 --redis=0 --system=0 >person.xml

Then the template can use the Oracle connection object :obj:`db` directly.


API
===
"""


import sys, os, argparse, datetime, keyword

from ll import ul4c, misc


__docformat__ = "reStructuredText"


class System:
	ul4attrs = {"execute"}

	def execute(self, cmd):
		return os.popen(cmd).read()


class Var:
	ul4attrs = {"value"}

	def __init__(self, value=None):
		self.value = value

	def ul4setattr(self, name, value):
		# As ``ul4attrs`` only contains ``"value"``, we will never be called with any other name
		self.value = value

	@misc.notimplemented
	def makevar(c, cursor):
		pass


class Connection:
	"""
	A :class:`Connection` object provides a database connection to an UL4
	template.

	To execute SQL the two methods :meth:`query` and :meth:`execute` are provided.

	Calling functions or procedures with out parameters can be done with variable
	objects that can be created with the methods :meth:`int`, :meth:`number`,
	:meth:`str`, :meth:`clob` and :meth:`date`. The resulting value of the out
	parameter is available from the :attr:`value` attribute of the variable
	object. The following example creates a function, calls it to get at the
	result and drops it again:

	.. sourcecode:: xml

		<?code db = oracle.connect('user/password@database')?>
		<?code db.execute('''
			create or replace function ul4test(p_arg integer)
			return integer
			as
			begin
				return 2*p_arg;
			end;
		''')?>
		<?code vout = db.int()?>
		<?code db.execute('begin ', vout, ' := ul4test(42); end;')?>
		<?print vout.value?>
		<?code db.execute('drop function ul4test')?>

	A :class:`Connection` object can be created with the methods
	:meth:`Globals.mysql` or :meth:`Globals.sqlite`.
	"""

	ul4attrs = {"query", "queryone", "execute", "int", "number", "str", "clob", "date"}

	def __init__(self, connection):
		self.connection = connection

	def _execute(self, cursor, queryparts):
		query = []
		params = {}
		vars = {}
		for (i, part) in enumerate(queryparts):
			if i % 2:
				name = "value{}".format((i+1)//2)
				if isinstance(part, Var):
					params[name] = part.makevar(cursor)
					vars[name] = part
				else:
					params[name] = part
				query.append(":" + name)
			else:
				query.append(part)
		cursor.execute("".join(query), **params)
		for (name, var) in vars.items():
			var.value = params[name].getvalue(0)

	def query(self, *queryparts):
		"""
		Execute the query passed in and return an iterator over the resulting
		records.

		At least one positional argument is required. Arguments alternate between
		fragments of the SQL query and parameters that will be embedded in the
		query. For example:

		.. sourcecode:: xml

			<?code db = globals.oracle("user/pwd@db")?>
			<?code name = "Bob"?>
			<ul>
				<?for p in db.query("select * from person where firstname=", name, " or lastname=", name)?>
					<li><?print p.firstname?> <?print p.lastname?></li>
				<?end for?>
			</ul>

		The records returned from :meth:`query` are dict-like objects mapping
		field names to field values.
		"""

		cursor = self.connection.cursor()
		self._execute(cursor, queryparts)
		return cursor

	def queryone(self, *queryparts):
		"""
		Execute the query passed in and return the first result record (or
		:const:`None` if the query didn't output any record). ``queryparts``
		is handled the same way as :meth:`query` does.
		"""

		cursor = self.connection.cursor()
		self._execute(cursor, queryparts)
		return cursor.fetchone()

	def execute(self, *queryparts):
		"""
		Similar to :meth:`query` and :meth:`queryone`, but doesn't doesn't return
		a result. This can be used to call functions or procedures.
		"""
		cursor = self.connection.cursor()
		self._execute(cursor, queryparts)

	@misc.notimplemented
	def str(self, value=None):
		"""
		Create a variable that can be used for OUT parameters of type ``varchar``.
		"""

	@misc.notimplemented
	def clob(self, value=None):
		"""
		Create a variable that can be used for OUT parameters of type ``clob``.
		"""

	@misc.notimplemented
	def int(self, value=None):
		"""
		Create a variable that can be used for OUT parameters of type ``integer``.
		"""

	@misc.notimplemented
	def number(self, value=None):
		"""
		Create a variable that can be used for OUT parameters of type ``number``.
		"""

	@misc.notimplemented
	def date(self, value=None):
		"""
		Create a variable that can be used for OUT parameters of type ``date``.
		"""


class OracleConnection(Connection):
	"""
	:class:`OracleConnection` is a subclass of :class:`Connection` that
	implements functionality that is specific to Oracle databases (e.g.
	support for variables). The inferface is the same as :class:`Connection`\s.

	An :class:`OracleConnection` object can be created with the method
	:meth:`Globals.oracle`.
	"""

	class IntVar(Var):
		def makevar(self, c):
			var = c.var(int)
			var.setvalue(0, self.value)
			return var

	class NumberVar(Var):
		def makevar(self, c):
			var = c.var(float)
			var.setvalue(0, self.value)
			return var

	class StrVar(Var):
		def makevar(self, c):
			var = c.var(str)
			var.setvalue(0, self.value)
			return var

	class CLOBVar(Var):
		def makevar(self, c):
			from ll import orasql
			var = c.var(orasql.CLOB)
			var.setvalue(0, self.value)
			return var

	class DateVar(Var):
		def makevar(self, c):
			var = c.var(datetime.datetime)
			var.setvalue(0, self.value)
			return var

	def __repr__(self):
		connectstring = "{}@{}".format(self.connection.username, self.connection.tnsentry)
		return "<{}.{} schema={!r} at {:#x}>".format(self.__class__.__module__, self.__class__.__qualname__, connectstring, id(self))

	def str(self, value=None):
		return self.StrVar(value)

	def clob(self, value=None):
		return self.CLOBVar(value)

	def int(self, value=None):
		return self.IntVar(value)

	def number(self, value=None):
		return self.NumberVar(value)

	def date(self, value=None):
		return self.DateVar(value)


class RedisConnection:
	"""
	A connection to a Redis database. A :class:`RedisConnection` object provides
	the methods :meth:`get` to read data from the database and :meth:`set` to
	write data to the database.

	Example::

		<?code db = redis.connect("192.168.123.42/1")?>
		<?code value = db.get("key")?>
		<?if isnone(value)?>
			<?code value = "foobar"?>
			<?code db.put("key", value, timedelta(seconds=10*60))?>
		<?end if?>
	"""

	ul4attrs = {"get", "put"}

	def __init__(self, host, port, db):
		import redis
		self.connection = redis.StrictRedis(host=host, port=port, db=db, decode_responses=True)

	def get(self, key):
		"""
		Return the value for the key ``key`` or ``None`` if the key doesn't exist.
		"""
		return self.connection.get(key)

	def set(self, key, data, timeout=None):
		"""
		Store the string value ``data`` under the key ``key``.

		If ``timeout`` is :const:`None` the value will be stored indefinitely.
		Otherwise it specifies when the value will expire. ``timeout`` can be
		an integer (the number of seconds) or a :class:`timedelta` object.
		"""
		if timeout is None:
			self.connection.set(key, data)
		else:
			self.connection.setex(key, timeout, data)


def fixname(name):
	newname = "".join(c for (i, c) in enumerate(name) if (c.isalnum() if i else c.isalpha()) or c == "_")
	while keyword.iskeyword(newname):
		newname += "_"
	return newname


def print_exception_chain(exc):
	print("UL4 traceback (most recent call last):", file=sys.stderr)
	for exc in misc.exception_chain(exc):
		print()
		print(misc.format_exception(exc), file=sys.stderr)


class Globals:
	"""
	An instance of the :class:`Globals` class will be passed to the main template
	as the ``globals`` variable. The following attributes will be accessible to
	UL4 templates:

	``templates`` : dictionary
		A dictionary containing the templates specified on the command line. This
		will include the main template.

	``vars`` : dictionary
		A dictionary containing the variables that have been specified via the
		:option:`-D`/:option:`--define` option.

	``encoding`` : string
		The encoding that will be used for output (this is the same as
		``sys.stdout.encoding``, so it can be set with the environment variable
		:envvar:`PYTHONIOENCODING`).

	``env``: dictionary
		A reference to :obj:`os.environ`.

	Furthermore the following methods can be called from UL4 templates:
	:meth:`error`, :meth:`log`, :meth:`oracle`, :meth:`mysql`, :meth:`sqlite`,
	:meth:`redis`, :meth:`system`, :meth:`load`, :meth:`save` and :meth:`compile`.
	"""

	ul4attrs = {"templates", "vars", "encoding", "env", "oracle", "mysql", "sqlite", "redis", "error", "log", "system", "load", "save", "compile"}
	def __init__(self, templates=None, vars=None, encoding=None, oracle=True, mysql=True, sqlite=True, redis=True, system=True, load=True, save=True, compile=True):
		self.templates = templates if templates is not None else {}
		self.encoding = encoding if encoding is not None else sys.stdout.encoding
		self.vars = vars if vars is not None else {}
		self.env = os.environ
		# Deactivate features if requested by overwriting the method with an instance attribute.
		if not oracle:
			self.oracle = None
		if not mysql:
			self.mysql = None
		if not sqlite:
			self.sqlite = None
		if not redis:
			self.redis = None
		if not system:
			self.system = None
		if not load:
			self.load = None
		if not save:
			self.save = None
		if not compile:
			self.compile = None

	def from_args(self, args):
		"""
		Sets the attributes of :obj:`self` from the object :obj:`args` (which
		must be an instance of :class:`argparse.Namespace`).

		Returns the main template.
		"""
		templates = {}
		maintemplate = None
		for templatename in args.templates:
			if templatename == "-":
				templatesource = sys.stdin.read()
				templatename = "stdin"
			else:
				with open(templatename, "r", encoding=args.encoding) as f:
					templatesource = f.read()
				templatename = os.path.basename(templatename)
				if os.path.extsep in templatename:
					templatename = templatename.rpartition(os.extsep)[0]
			templatename = fixname(templatename)
			if args.stacktrace == "short":
				try:
					template = ul4c.Template(templatesource, name=templatename, whitespace=args.whitespace)
				except Exception as exc:
					print_exception_chain(exc)
					raise SystemExit(1)
			else:
				template = ul4c.Template(templatesource, name=templatename, whitespace=args.whitespace)
			# The first template is the main template
			if maintemplate is None:
				maintemplate = template
			templates[template.name] = template
		self.templates = templates

		self.vars = dict(args.vars) if args.vars is not None else {}

		def option(name):
			if getattr(args, name):
				if getattr(self, name) is None:
					delattr(self, name)
			else:
				if getattr(self, name) is not None:
					setattr(self, name, None)

		option("oracle")
		option("mysql")
		option("sqlite")
		option("redis")
		option("system")
		option("load")
		option("save")
		option("compile")

		return maintemplate

	def error(self, message, ast=None):
		"""
		Can be called to output an error message and abort template execution.
		The signature is:

		.. sourcecode:: python

			globals.error(message, ast=None)

		``message`` is the error message and ``ast`` can be an AST node from an
		UL4 template syntax tree to print an error message that originates from
		that node.
		"""
		exc = Exception(message)
		if ast is not None:
			raise ul4c.Error(ast) from exc
		else:
			raise exc

	def log(self, *args):
		"""
		Logs ``args`` to ``sys.stderr``.
		"""
		print(*args, file=sys.stderr)

	def oracle(self, connectstring):
		"""
		Return an :class:`OracleConnection` object for the Oracle connect string
		passed in::

			<?code db = globals.oracle("user/password@database")?>
			<?for row in db.query("select sysdate as sd from dual")?>
				<?print row.sd?>
			<?end for?>
		"""
		from ll import orasql
		return OracleConnection(orasql.connect(connectstring, readlobs=True))

	def mysql(self, connectstring):
		"""
		Return a :class:`Connection` object to a MySQL database for the
		connectstring passed in. The format of the connect string is::

			user/password@host/database
		"""
		import MySQLdb
		from MySQLdb import cursors
		(user, host) = connectstring.split("@")
		(user, passwd) = user.split("/")
		(host, db) = host.split("/")
		return Connection(MySQLdb.connect(user=user, passwd=passwd, host=host, db=db, use_unicode=True, cursorclass=cursors.DictCursor))

	def sqlite(self, connectstring):
		"""
		Return a :class:`Connection` object to an SQLite database for the
		connectstring passed in. The connectstring will be passed directly
		to :func:`sqlite3.connect`.
		"""
		import sqlite3
		connection = sqlite3.connect(connectstring)
		class Row(sqlite3.Row):
			def __getitem__(self, key):
				if isinstance(key, str):
					key = key.encode("ascii")
				return sqlite3.Row.__getitem__(self, key)
		connection.row_factory = Row
		return Connection(connection)

	def redis(self, connectstring):
		"""
		Return a :class:`RedisConnection` object, which provides a connection to a
		Redis database. The connectstring has the format::

			host:port/db

		``port`` is optional and defaults to 6379. ``db`` is optional too and
		defaults to 0.
		"""
		(hostport, _, db) = connectstring.partition("/")
		if not db:
			db = 0
		(host, _, port) = hostport.partition(":")
		if not port:
			port = 6379
		return RedisConnection(host=host, port=port, db=db)

	def system(self, cmd):
		"""
		Execute the system command :obj:`cmd` and returns its output, e.g.
		the template:

		.. sourcecode:: xml

			<?print globals.system("whoami")?>

		will output the user name.
		"""
		return os.popen(cmd).read()

	def load(self, filename, encoding="utf-8"):
		"""
		Read a file from disk and returns the content. :obj:`filename` is the
		filename and :obj:`encoding` is the encoding of the file. The encoding
		parameter is optional and defaults to ``"utf-8"``:

		.. sourcecode:: xml

			<?code data = globals.load("/home/user/data.txt", "iso-8859-1")?>
		"""
		with open(filename, "r", encoding=encoding) as f:
			return f.read()

	def save(self, filename, data, encoding="utf-8"):
		r"""
		Save the string :obj:`data` to a file on disk. :obj:`filename` is the
		filename and :obj:`encoding` is the encoding of the file. The encoding
		parameter is optional and defaults to ``"utf-8"``:

		.. sourcecode:: xml

			<?code globals.save("/home/user/data.txt", "foo\nbar\n", "iso-8859-1")?>
		"""
		with open(filename, "w", encoding=encoding) as f:
			f.write(data)

	def compile(self, source, name=None, whitespace="keep", signature=None, startdelim="<?", enddelim="?>"):
		"""
		Compile the UL4 source ``source`` into a :class:`~ll.ul4c.Template` object
		and return it. All other parameters are passed to the
		:class:`~ll.ul4c.Template` constructor too.
		"""
		return ul4c.Template(source, name=name, whitespace=whitespace, signature=signature, startdelim=startdelim, enddelim=enddelim)

	def define(self, arg):
		(name, _, value) = arg.partition("=")
		(name, _, type) = name.partition(":")
		if any(c != "_" and not (c.isalnum() if i else c.isalpha()) for (i, c) in enumerate(name)):
			raise argparse.ArgumentTypeError("{!r} is not a legal variable name".format(name))

		if type == "int":
			if not value:
				return (name, 0)
			try:
				return (name,  int(value))
			except ValueError:
				raise argparse.ArgumentTypeError("{!r} is not a legal integer value".format(value))
		elif type == "float":
			if not value:
				return (name, 0.)
			try:
				return (name, float(value))
			except ValueError:
				raise argparse.ArgumentTypeError("{!r} is not a legal float value".format(value))
		elif type == "bool":
			if value in ("", "0", "no", "false", "False"):
				return (name, False)
			if value in ("1", "yes", "true", "True"):
				return (name, True)
			raise argparse.ArgumentTypeError("{!r} is not a legal bool value".format(value))
		elif type == "oracle":
			return (name, self.oracle(value))
		elif type == "sqlite":
			return (name, self.sqlite(value))
		elif type == "mysql":
			return (name, self.mysql(value))
		elif type == "redis":
			return (name, self.redis(value))
		elif type and type != "str":
			raise argparse.ArgumentTypeError("{!r} is not a legal type".format(type))
		return (name, value)


def main(args=None):
	globals = Globals()

	define = globals.define

	p = argparse.ArgumentParser(description="render UL4 templates with access to Oracle, MySQL, SQLite or Redis databases", epilog="For more info see http://www.livinglogic.de/Python/scripts_rul4.html")
	p.add_argument("templates", metavar="template", help="templates to be used (first template gets rendered)", nargs="+")
	p.add_argument("-e", "--encoding", dest="encoding", help="Encoding for template sources (default %(default)s)", default="utf-8", metavar="ENCODING")
	p.add_argument("-w", "--whitespace", dest="whitespace", help="How to treat whitespace in template sources? (default %(default)s)", choices=("keep", "strip", "smart"), default="smart")
	p.add_argument("-t", "--stacktrace", dest="stacktrace", help="How to display stack traces in case of an error? (default %(default)s)", choices=("full", "short"), default="short")
	p.add_argument(      "--oracle", dest="oracle", help="Allow the templates to connect to Oracle databases? (default %(default)s)", action=misc.FlagAction, default=True)
	p.add_argument(      "--sqlite", dest="sqlite", help="Allow the templates to connect to SQLite databases? (default %(default)s)", action=misc.FlagAction, default=True)
	p.add_argument(      "--mysql", dest="mysql", help="Allow the templates to connect to MySQL databases? (default %(default)s)", action=misc.FlagAction, default=True)
	p.add_argument(      "--redis", dest="redis", help="Allow the templates to connect to Redis databases? (default %(default)s)", action=misc.FlagAction, default=True)
	p.add_argument(      "--system", dest="system", help="Allow the templates to execute system commands? (default %(default)s)", action=misc.FlagAction, default=True)
	p.add_argument(      "--load", dest="load", help="Allow the templates to load data from arbitrary paths? (default %(default)s)", action=misc.FlagAction, default=True)
	p.add_argument(      "--save", dest="save", help="Allow the templates to save data to arbitrary paths? (default %(default)s)", action=misc.FlagAction, default=True)
	p.add_argument(      "--compile", dest="compile", help="Allow the templates access to the compile function? (default %(default)s)", action=misc.FlagAction, default=True)
	p.add_argument("-D", "--define", dest="vars", metavar="var=value", help="Pass additional parameters to the template (can be specified multiple times).", action="append", type=define)

	args = p.parse_args(args)

	maintemplate = globals.from_args(args)

	if args.stacktrace == "short":
		try:
			for part in maintemplate.render(globals=globals):
				sys.stdout.write(part)
		except Exception as exc:
			print_exception_chain(exc)
			return 1
	else:
		for part in maintemplate.render(globals=globals):
			sys.stdout.write(part)


if __name__ == "__main__":
	sys.exit(main())



# -*- coding: utf-8 -*-
# cython: language_level=3, always_allow_keywords=True

## Copyright 2011-2016 by LivingLogic AG, Bayreuth/Germany
## Copyright 2011-2016 by Walter Dörwald
##
## All Rights Reserved
##
## See ll/xist/__init__.py for the license


"""
This module provides functions for encoding and decoding a lightweight
text-based format for serializing the object types supported by UL4.

It is extensible to allow encoding/decoding arbitrary instances (i.e. it is
basically a reimplementation of :mod:`pickle`, but with string input/output
instead of bytes and with an eye towards cross-plattform support).

There are implementations for Python (this module), Java_ and Javascript_
(as part of the UL4 packages for those languages).

.. _Java: https://github.com/LivingLogic/LivingLogic.Java.ul4
.. _Javascript: https://github.com/LivingLogic/LivingLogic.Javascript.ul4

Furthermore there's an `Oracle package`_ that can be used for generating
UL4ON encoded data.

.. _Oracle package: https://github.com/LivingLogic/LivingLogic.Oracle.ul4

Basic usage follows the API design of :mod:`pickle`, :mod:`json`, etc. and
supports most builtin Python types::

	>>> from ll import ul4on
	>>> ul4on.dumps(None)
	'n'
	>>> ul4on.loads('n')
	>>> ul4on.dumps(False)
	'bF'
	>>> ul4on.loads('bF')
	False
	>>> ul4on.dumps(42)
	'i42'
	>>> ul4on.loads('i42')
	42
	>>> ul4on.dumps(42.5)
	'f42.5'
	>>> ul4on.loads('f42.5')
	42.5
	>>> ul4on.dumps('foo')
	"S'foo'"
	>>> ul4on.loads("S'foo'")
	'foo'

:class:`datetime` and :class:`timedelta` objects are supported too::

	>>> import datetime
	>>> ul4on.dumps(datetime.datetime.now())
	'Z i2014 i11 i3 i18 i16 i45 i314157'
	>>> ul4on.loads('Z i2014 i11 i3 i18 i16 i45 i314157')
	datetime.datetime(2014, 11, 3, 18, 16, 45, 314157)
	>>> ul4on.dumps(datetime.timedelta(days=1))
	'T i1 i0 i0'
	>>> ul4on.loads('T i1 i0 i0')
	datetime.timedelta(1)

:mod:`ll.ul4on` also supports :class:`Color` objects from :mod:`ll.color`::

	>>> from ll import color
	>>> ul4on.dumps(color.red)
	'C i255 i0 i0 i255'
	>>> ul4on.loads('C i255 i0 i0 i255')
	Color(0xff, 0x00, 0x00)

Lists, dictionaries and sets are also supported::

	>>> ul4on.dumps([1, 2, 3])
	'L i1 i2 i3 ]'
	>>> ul4on.loads('L i1 i2 i3 ]')
	[1, 2, 3]
	>>> ul4on.dumps(dict(one=1, two=2))
	"D S'two' i2 S'one' i1 }"
	>>> ul4on.loads("D S'two' i2 S'one' i1 }")
	{'one': 1, 'two': 2}
	>>> ul4on.dumps({1, 2, 3})
	'Y i1 i2 i3 }'
	>>> ul4on.loads('Y i1 i2 i3 }')
	{1, 2, 3}

:mod:`ll.ul4on` can also handle recursive data structures::

	>>> r = []
	>>> r.append(r)
	>>> ul4on.dumps(r)
	'L ^0 ]'
	>>> r2 = ul4on.loads('L ^0 ]')
	>>> r2
	[[...]]
	>>> r2 is r2[0]
	True
	>>> r = {}
	>>> r['recursive'] = r
	>>> ul4on.dumps(r)
	"D S'recursive' ^0 }"
	>>> r2 = ul4on.loads("D S'recursive' ^0 }")
	>>> r2
	{'recursive': {...}}
	>>> r2['recursive'] is r2
	True

UL4ON is extensible. It supports serializing arbitrary instances by registering
the class with the UL4ON serialization machinery::

	from ll import ul4on

	@ul4on.register("com.example.person")
	class Person:
		def __init__(self, firstname=None, lastname=None):
			self.firstname = firstname
			self.lastname = lastname

		def __repr__(self):
			return "<Person firstname={!r} lastname={!r}>".format(self.firstname, self.lastname)

		def ul4ondump(self, encoder):
			encoder.dump(self.firstname)
			encoder.dump(self.lastname)

		def ul4onload(self, decoder):
			self.firstname = decoder.load()
			self.lastname = decoder.load()

	jd = Person("John", "Doe")
	output = ul4on.dumps(jd)
	print("Dump:", output)
	jd2 = ul4on.loads(output)
	print("Loaded:", jd2)

This script outputs::

	Dump: O S'com.example.person' S'John' S'Doe' )
	Loaded: <Person firstname='John' lastname='Doe'>
"""

import datetime, collections, io, ast


__docformat__ = "reStructuredText"


_registry = {}


def register(name):
	"""
	This decorator can be used to register the decorated class with the
	:mod:`ll.ul4on` serialization machinery.

	:obj:`name` must be a globally unique name for the class. To avoid
	name collisions Java's class naming system should be used (i.e. an
	inverted domain name like ``com.example.foo.bar``).

	:obj:`name` will be stored in the class attribute ``ul4onname``.
	"""
	def registration(cls):
		cls.ul4onname = name
		_registry[name] = cls
		return cls
	return registration


class Encoder:
	def __init__(self, stream, indent=None):
		"""
		Create an encoder for serializing objects to  :obj:`self.stream`.

		:obj:`stream` must provide a :meth:`write` method.
		"""
		self.stream = stream
		self._level = 0
		self.indent = indent
		self._lastwaslf = False
		self._first = True # Remember whether we have dumped something into the stream (so we have to write separator whitespace/indentation) or not
		self._objects = []
		self._id2index = {}

	def _record(self, obj):
		# Record that we've written this object and in which position
		self._id2index[id(obj)] = len(self._objects)
		self._objects.append(obj)

	def _line(self, line, *items):
		if self.indent:
			self.stream.write(self.indent*self._level)
		else:
			if not self._first:
				self.stream.write(" ")
		self._first = False
		self.stream.write(line)
		if items:
			oldindent = self.indent
			try:
				self.indent = ""
				for item in items:
					self.dump(item)
			finally:
				self.indent = oldindent
		if self.indent:
			self.stream.write("\n")

	def dump(self, obj):
		"""
		Serialize :obj:`obj` as an UL4ON formatted stream.
		"""
		# Have we written this object already?
		if id(obj) in self._id2index:
			# Yes: Store a backreference to the object
			self._line("^{}".format(self._id2index[id(obj)]))
		else:
			from ll import ul4c, color, misc
			# No: Write the object itself
			# We're not using backreferences if the object itself has a shorter dump
			if obj is None:
				self._line("n")
			elif isinstance(obj, bool):
				self._line("bT" if obj else "bF")
			elif isinstance(obj, int):
				self._line("i{}".format(obj))
			elif isinstance(obj, float):
				self._line("f{!r}".format(obj))
			elif isinstance(obj, str):
				self._record(obj)
				self._line("S{!r}".format(obj))
			elif isinstance(obj, slice):
				self._record(obj)
				self._line("R", obj.start, obj.stop)
			elif isinstance(obj, color.Color):
				self._record(obj)
				self._line("C", obj.r(), obj.g(), obj.b(), obj.a())
			elif isinstance(obj, datetime.datetime):
				self._record(obj)
				self._line("Z", obj.year, obj.month, obj.day, obj.hour, obj.minute, obj.second, obj.microsecond)
			elif isinstance(obj, datetime.date):
				self._record(obj)
				self._line("Z", obj.year, obj.month, obj.day, 0, 0, 0, 0)
			elif isinstance(obj, datetime.timedelta):
				self._record(obj)
				self._line("T", obj.days, obj.seconds, obj.microseconds)
			elif isinstance(obj, misc.monthdelta):
				self._record(obj)
				self._line("M", obj.months())
			elif isinstance(obj, collections.Sequence):
				self._record(obj)
				self._line("L")
				self._level += 1
				for item in obj:
					self.dump(item)
				self._level -= 1
				self._line("]")
			elif isinstance(obj, collections.Mapping):
				self._record(obj)
				self._line("E" if isinstance(obj, collections.OrderedDict) else "D")
				self._level += 1
				for (key, item) in obj.items():
					self.dump(key)
					self.dump(item)
				self._level -= 1
				self._line("}")
			elif isinstance(obj, collections.Set):
				self._record(obj)
				self._line("Y")
				self._level += 1
				for item in obj:
					self.dump(item)
				self._level -= 1
				self._line("}")
			else:
				self._record(obj)
				self._line("O", obj.ul4onname)
				self._level += 1
				obj.ul4ondump(self)
				self._level -= 1
				self._line(")")


class Decoder:
	def __init__(self, stream):
		"""
		Create a decoder for deserializing objects from  :obj:`self.stream`.

		:obj:`stream` must provide a :meth:`read` method.
		"""
		self.stream = stream
		self._objects = []
		self._keycache = {} # Used for "interning" dictionary keys

	def load(self):
		"""
		Deserialize the next object in the stream and return it.
		"""
		return self._load(None)

	def _readint(self):
		buffer = []
		while True:
			c = self.stream.read(1)
			if c and not c.isspace():
				buffer.append(c)
			else:
				return int("".join(buffer))

	def _loading(self, obj):
		self._objects.append(obj)

	def _nextchar(self, nextchar=None):
		while True:
			nextchar = self.stream.read(1)
			if nextchar:
				if not nextchar.isspace():
					return nextchar
			else:
				raise EOFError()

	def _beginfakeloading(self):
		# For loading custom object or immutable objects that have attributes we have a problem:
		# We have to record the object we're loading *now*, so that it is available for backreferences.
		# However until we've read the UL4ON name of the class (for custom object) or the attributes
		# of the object (for immutable objects with attributes), we can't create the object.
		# So we push ``None`` to the backreference list for now and put the right object in this spot,
		# once we've created it (via :meth:`_endfakeloading`). This shouldn't lead to problems,
		# because during the time the backreference is wrong, only the class name is read,
		# so our object won't be referenced. For immutable objects the attributes normally
		# don't reference the object itself.
		oldpos = len(self._objects)
		self._loading(None)
		return oldpos

	def _endfakeloading(self, oldpos, value):
		# Fix backreference in object list
		self._objects[oldpos] = value

	def _load(self, typecode):
		from ll import misc
		if typecode is None:
			typecode = self._nextchar()
		if typecode == "^":
			position = self._readint()
			return self._objects[position]
		elif typecode in "nN":
			if typecode == "N":
				self._loading(None)
			return None
		elif typecode in "bB":
			value = self.stream.read(1)
			if value == "T":
				value = True
			elif value == "F":
				value = False
			else:
				raise ValueError("broken UL4ON stream at position {}: expected 'T' or 'F' for bool; got {!r}".format(self.stream.tell(), value))
			if typecode == "B":
				self._loading(value)
			return value
		elif typecode in "iI":
			value = self._readint()
			if typecode == "I":
				self._loading(value)
			return value
		elif typecode in "fF":
			chars = []
			while True:
				c = self.stream.read(1)
				if c and not c.isspace():
					chars.append(c)
				else:
					value = float("".join(chars))
					break
			if typecode == "F":
				self._loading(value)
			return value
		elif typecode in "sS":
			delimiter = self.stream.read(1)
			if not delimiter:
				raise EOFError()
			buffer = [delimiter]
			while True:
				c = self.stream.read(1)
				if not c:
					raise EOFError()
				buffer.append(c)
				if c == delimiter:
					value = ast.literal_eval("".join(buffer))
					break
				elif c == "\\":
					c2 = self.stream.read(1)
					if not c2:
						raise EOFError()
					buffer.append(c2)
			if typecode == "S":
				self._loading(value)
			return value
		elif typecode in "cC":
			from ll import color
			if typecode == "C":
				oldpos = self._beginfakeloading()
			r = self._load(None)
			g = self._load(None)
			b = self._load(None)
			a = self._load(None)
			value = color.Color(r, g, b, a)
			if typecode == "C":
				self._endfakeloading(oldpos, value)
			return value
		elif typecode in "zZ":
			if typecode == "Z":
				oldpos = self._beginfakeloading()
			year = self._load(None)
			month = self._load(None)
			day = self._load(None)
			hour = self._load(None)
			minute = self._load(None)
			second = self._load(None)
			microsecond = self._load(None)
			value = datetime.datetime(year, month, day, hour, minute, second, microsecond)
			if typecode == "Z":
				self._endfakeloading(oldpos, value)
			return value
		elif typecode in "rR":
			if typecode == "R":
				oldpos = self._beginfakeloading()
			start = self._load(None)
			stop = self._load(None)
			value = slice(start, stop)
			if typecode == "R":
				self._endfakeloading(oldpos, value)
			return value
		elif typecode in "tT":
			if typecode == "T":
				oldpos = self._beginfakeloading()
			days = self._load(None)
			seconds = self._load(None)
			microseconds = self._load(None)
			value = datetime.timedelta(days, seconds, microseconds)
			if typecode == "T":
				self._endfakeloading(oldpos, value)
			return value
		elif typecode in "mM":
			from ll import misc
			if typecode == "M":
				oldpos = self._beginfakeloading()
			months = self._load(None)
			value = misc.monthdelta(months)
			if typecode == "M":
				self._endfakeloading(oldpos, value)
			return value
		elif typecode in "lL":
			value = []
			if typecode == "L":
				self._loading(value)
			while True:
				typecode = self._nextchar()
				if typecode == "]":
					return value
				else:
					item = self._load(typecode)
					value.append(item)
		elif typecode in "dDeE":
			value = {} if typecode in "dD" else collections.OrderedDict()
			if typecode in "DE":
				self._loading(value)
			while True:
				typecode = self._nextchar()
				if typecode == "}":
					return value
				else:
					key = self._load(typecode)
					if isinstance(key, str):
						if key in self._keycache:
							key = self._keycache[key]
						else:
							self._keycache[key] = key
					item = self._load(None)
					value[key] = item
		elif typecode in "yY":
			value = set()
			if typecode == "Y":
				self._loading(value)
			while True:
				typecode = self._nextchar()
				if typecode == "}":
					return value
				else:
					item = self._load(typecode)
					value.add(item)
		elif typecode in "oO":
			if typecode == "O":
				oldpos = self._beginfakeloading()
			name = self._load(None)
			try:
				cls = _registry[name]
			except KeyError:
				raise TypeError("can't decode object of type {}".format(name))
			value = cls()
			if typecode == "O":
				self._endfakeloading(oldpos, value)
			value.ul4onload(self)
			typecode = self._nextchar()
			if typecode != ")":
				raise ValueError("broken UL4ON stream at position {}: object terminator ')' expected, got {!r}".format(self.stream.tell(), typecode))
			return value
		else:
			raise ValueError("broken UL4ON stream at position {}: unknown typecode {!r}".format(self.stream.tell(), typecode))


class StreamBuffer:
	# Internal helper class that wraps a file-like object and provides buffering
	def __init__(self, stream, bufsize=1024*1024):
		self.stream = stream
		self.bufsize = bufsize
		self.buffer = ""

	def read(self, size):
		havesize = len(self.buffer)
		if havesize >= size:
			result = self.buffer[:size]
			self.buffer = self.buffer[size:]
			return result
		else:
			needsize = size-havesize
			newdata = self.stream.read(max(self.bufsize, needsize))
			result = self.buffer + newdata[:needsize]
			self.buffer = newdata[needsize:]
			return result


def dumps(obj, indent=None):
	"""
	Serialize :obj:`obj` as an UL4ON formatted string.
	"""
	stream = io.StringIO()
	Encoder(stream, indent=indent).dump(obj)
	return stream.getvalue()


def dump(obj, stream, indent=None):
	"""
	Serialize :obj:`obj` as an UL4ON formatted stream to :obj:`stream`.

	:obj:`stream` must provide a :meth:`write` method.
	"""
	Encoder(stream, indent=indent).dump(obj)


def loadclob(clob, bufsize=1024*1024):
	"""
	Deserialize :obj:`clob` (which must be an :mod:`cx_Oracle` ``CLOB`` variable
	containing an UL4ON formatted object) to a Python object.

	:obj:`bufsize` specifies the chunk size for reading the underlying ``CLOB``
	object.
	"""
	return Decoder(StreamBuffer(clob, bufsize)).load()


def loads(string):
	"""
	Deserialize :obj:`string` (which must be a string containing an UL4ON
	formatted object) to a Python object.
	"""
	return Decoder(io.StringIO(string)).load()


def load(stream):
	"""
	Deserialize :obj:`stream` (which must be file-like object with a :meth:`read`
	method containing an UL4ON formatted object) to a Python object.
	"""
	return Decoder(stream).load()



#! /usr/bin/env/python
# -*- coding: utf-8 -*-
# cython: language_level=3, always_allow_keywords=True

## Copyright 1999-2016 by LivingLogic AG, Bayreuth/Germany
## Copyright 1999-2016 by Walter Dörwald
##
## All Rights Reserved
##
## See ll/xist/__init__.py for the license


import pytest

from ll import misc
from ll.xist import xsc, xfind, parse
from ll.xist.ns import html, php, abbr


node = xsc.Frag(
	html.div(
		html.h1("The ", html.em("important"), " headline"),
		html.p("The ", html.em("first"), " paragraph."),
		html.p("The ", html.em("second"), " ", html.em("important"), " paragraph."),
		align="left",
	),
	html.div(
		html.h1("The headline", html.img(src="root:gurk.gif")),
		html.p("The ", html.em("first"), " paragraph."),
		html.div(
			html.h2("The ", html.em("important"), " headline"),
			html.p("The ", html.em("second"), " ", html.em("important"), " paragraph."),
			id="id42",
		),
		html.div(id="id23"),
		class_="foo",
	),
	html.p(html.em("only")),
	php.php("echo $footer;"),
	abbr.xist(),
)


def test_levels():
	def check(node, expr, ids):
		assert "".join(str(e.attrs.id) for e in node.walknodes(expr)) == ids

	ds = [html.div(id=id) for id in range(8)]
	ds[1].append(ds[4:7])
	ds[2].append(ds[7])
	ds[0].append(ds[1:4])
	#      ____0____
	#     /    |    \
	#   _1_    2     3
	#  / | \   |
	# 4  5  6  7

	tests = [
		(ds[0]//html.div, "1456273"),
		(ds[0]/html.div, "123"),
		(ds[0]/html.div/html.div, "4567"),
		(ds[0]/html.div/html.div/html.div, ""),
	]
	for (got, exp) in tests:
		check(ds[0], got, exp)


def test_isinstance():
	res = list(node.walknodes(html.h1))
	assert len(res) == 2
	assert res[0] is node[0][0]
	assert res[1] is node[1][0]

	def check(res):
		assert len(res) == 3
		assert res[0] is node[0][0]
		assert res[1] is node[1][0]
		assert res[2] is node[1][2][0]

	check(list(node.walknodes(xfind.IsInstanceSelector(html.h1, html.h2))))
	check(list(node.walknodes(xfind.IsInstanceSelector(html.h1) | html.h2)))
	check(list(node.walknodes(html.h1 | xfind.IsInstanceSelector(html.h2))))
	check(list(node.walknodes(html.h1 | html.h2)))
	check(list(node.walknodes(html.h1, html.h2)))
	check(list(node.walknodes(html.h1, html.h2, ~xfind.any)))
	check(list(node.walknodes(xsc.Element & ~(xsc.Text | html.p | html.div | html.em | html.img))))


def test_element():
	def check(expr, res):
		assert [str(e) for e in node.walknodes(expr)] == res
	result = ["important", "first", "second", "important", "first", "important", "second", "important", "only"]
	check(xfind.element(None, "gurk"), [])
	check(xfind.element("nix", "gurk"), [])
	check(xfind.element(html, "em"), result)
	check(xfind.element(html.xmlns, "em"), result)


def test_procinst():
	def check(expr, res):
		assert [e for e in node.walknodes(expr)] == res
	check(xfind.procinst("foo"), [])
	check(xfind.procinst("php"), [node[-2]])


def test_entity():
	def check(expr, res):
		assert [e for e in node.walknodes(expr)] == res
	check(xfind.entity("foo"), [])
	check(xfind.entity("xist"), [node[-1]])


def test_is():
	# Frags will be put into the path (but only as the root of the path),
	# but the selector will not be called for the Frag, so when the first call
	# happens there are already two nodes in the path. This is done on purpose:
	# selectors should not have to special case Frags
	res = list(node.walknodes(node))
	assert len(res) == 0

	res = list(node.walknodes(node[0]))
	assert len(res) == 1
	assert res[0] is node[0]


def test_isroot():
	res = list(node.walknodes(xfind.isroot))
	assert len(res) == 0

	res = list(node[0].walknodes(xfind.isroot))
	assert len(res) == 1
	assert res[0] is node[0]


def test_empty():
	res = list(node.walknodes(xfind.empty))
	assert len(res) == 2
	assert res[0] is node[1][0][-1]
	assert res[1] is node[1][-1]


def test_onlychild():
	res = list(node.walknodes(xfind.onlychild & html.em))
	assert len(res) == 1
	assert res[0] is node[2][0]


def test_onlyoftype():
	res = list(node.walknodes(xfind.onlyoftype & html.h1))
	assert len(res) == 2
	assert res[0] is node[0][0]
	assert res[1] is node[1][0]

	res = list(node.walknodes(xfind.onlyoftype & html.div))
	assert len(res) == 0

	res = list(node.walknodes(xfind.onlyoftype & html.p))
	assert len(res) == 3
	assert res[0] is node[1][1]
	assert res[1] is node[1][2][1]
	assert res[2] is node[2]


def test_hasattr():
	# hasattr
	res = list(node.walknodes(xfind.hasattr("class")))
	assert len(res) == 1
	assert res[0] is node[1]

	res = list(node.walknodes(xfind.hasattr(html.div.Attrs.id, html.div.Attrs.align)))
	assert len(res) == 3
	assert res[0] is node[0]
	assert res[1] is node[1][2]
	assert res[2] is node[1][3]


def test_attrhasvalue():
	def check(expected, attrname, *attrvalues):
		got = list(node.walknodes(xfind.attrhasvalue(attrname, *attrvalues)))
		assert len(got) == len(expected)
		for (gotnode, expectednode) in zip(got, expected):
			assert gotnode is expectednode

	check([node[0]], "align", "left")
	check([node[0]], html.div.Attrs.align, "left")
	check([node[0]], "align", "right", "center", "left")
	check([], "align", "right", "center")
	check([], "align", "right")
	check([], "gurk", "hurz")


def test_attrcontains():
	def check(expected, attrname, *attrvalues):
		got = list(node.walknodes(xfind.attrcontains(attrname, *attrvalues)))
		assert len(got) == len(expected)
		for (gotnode, expectednode) in zip(got, expected):
			assert gotnode is expectednode

	check([node[0]], "align", "ef")
	check([node[0]], html.div.Attrs.align, "ef")
	check([node[0]], "align", "ri", "ef")
	check([], "align", "ri", "en")
	check([], "align", "x")
	check([], "gurk", "nix")


def test_attrstartswith():
	def check(expected, attrname, *attrvalues):
		got = list(node.walknodes(xfind.attrstartswith(attrname, *attrvalues)))
		assert len(got) == len(expected)
		for (gotnode, expectednode) in zip(got, expected):
			assert gotnode is expectednode

	check([node[0]], "align", "le")
	check([node[0]], html.div.Attrs.align, "le")
	check([node[0]], "align", "ri", "ce", "le")
	check([], "align", "ri", "ce")
	check([], "align", "eft")
	check([], "gurk", "nix")
	check([node[1][0][1]], "src", "root:")


def test_attrendswith():
	def check(expected, attrname, *attrvalues):
		got = list(node.walknodes(xfind.attrendswith(attrname, *attrvalues)))
		assert len(got) == len(expected)
		for (gotnode, expectednode) in zip(got, expected):
			assert gotnode is expectednode

	check([node[0]], "align", "ft")
	check([node[0]], html.div.Attrs.align, "ft")
	check([node[0]], "align", "ht", "er", "ft")
	check([], "align", "ht", "er")
	check([], "align", "lef")
	check([], "gurk", "nix")
	check([node[1][0][1]], "src", ".gif")


def test_hasid():
	res = list(node.walknodes(xfind.hasid("id42")))
	assert len(res) == 1
	assert res[0] is node[1][2]


def test_hasclass():
	res = list(node.walknodes(xfind.hasclass("foo")))
	assert len(res) == 1
	assert res[0] is node[1]


def test_frag():
	e = parse.tree(b"das ist <b>klaus</b>. das ist <b>erich</b>", parse.SGMLOP(), parse.NS(html), parse.Node())
	assert "".join(map(str, e.walknodes(e//html.b))) == "klauserich"


def test_multiall():
	def check(node, expr, ids):
		assert "".join(str(e.attrs.id) for e in node.walknodes(expr)) == ids

	#        ____0____
	#       /         \
	#     _1_         _2_
	#    /   \       /   \
	#   3     4     5     6
	#  / \   / \   / \   / \
	# 7   8 9   a b   c d   e
	ds = [html.div(id=hex(id).lower()[2:]) for id in range(15)]
	for i in range(7):
		ds[i].append(ds[2*i+1:2*i+3])
	check(ds[0], ds[0]//html.div//html.div, "37849a5bc6de")


def test_itemsslices():
	def check(node, expr, ids):
		assert "".join(str(e.attrs.id) for e in node.walknodes(expr)) == ids

	#        ____0____
	#       /    |    \
	#     _1_   _2_   _3_
	#    /   \ /   \ /   \
	#   4     5     6     7
	ds = [html.div(id=id) for id in range(8)]
	ds[0].append(ds[1], ds[2], ds[3])
	ds[1].append(ds[4], ds[5])
	ds[2].append(ds[5], ds[6])
	ds[3].append(ds[6], ds[7])

	tests = [
		(ds[0]/html.div[0]/html.div[-1], "5"),
		(ds[0]/html.div/html.div[-1], "567"),
		(ds[0]/html.div[-1]/html.div, "67"),
		(ds[0]/html.div/html.div, "455667"), # we get 5 and 6 twice
		#(ds[0]/(html.div/html.div) & xfind.nthchild(2), u"5"), # we get 5 and 6 twice
		#(ds[0]/html.div[:]/html.div[:], u"455667"),
		(ds[0]/html.div/html.p[0], ""),
		(ds[0]/html.p[0]/html.p[0], ""),
		(ds[0]//html.div, "145256367"),
	]
	for (got, exp) in tests:
		check(ds[0], got, exp)


def test_item():
	e = html.div(range(10))
	assert str(misc.item(e[xsc.Text], 0)) == "0"
	assert str(misc.item(e[xsc.Text], 9)) == "9"
	assert str(misc.item(e[xsc.Text], -1)) == "9"
	assert str(misc.item(e[xsc.Text], -10)) == "0"
	misc.item(e[xsc.Text], 10) is None
	misc.item(e[xsc.Text], -11) is None
	assert str(misc.item(e[xsc.Text], 10, "x")) == "x"
	assert str(misc.item(e[xsc.Text], -11, "x")) == "x"



# -*- coding: utf-8 -*-
# cython: language_level=3, always_allow_keywords=True

## Copyright 2007-2016 by LivingLogic AG, Bayreuth/Germany
## Copyright 2007-2016 by Walter Dörwald
##
## All Rights Reserved
##
## See ll/xist/__init__.py for the license


r"""
This module can be used on UNIX to fork a daemon process. It is based on
`Jürgen Hermann's Cookbook recipe`__.

__ http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/66012

An example script might look like this::

	from ll import daemon

	counter = daemon.Daemon(
		stdin="/dev/null",
		stdout="/tmp/daemon.log",
		stderr="/tmp/daemon.log",
		pidfile="/var/run/counter/counter.pid",
		user="nobody"
	)

	if __name__ == "__main__":
		if counter.service():
			import sys, os, time
			sys.stdout.write("Daemon started with pid {}\n".format(os.getpid()))
			sys.stdout.write("Daemon stdout output\n")
			sys.stderr.write("Daemon stderr output\n")
			c = 0
			while True:
				sys.stdout.write('{}: {}\n'.format(c, time.ctime(time.time())))
				sys.stdout.flush()
				c += 1
				time.sleep(1)
"""


import sys, os, signal, pwd, grp, argparse


__docformat__ = "reStructuredText"


class Daemon:
	"""
	The :class:`Daemon` class provides methods for starting and stopping a
	daemon process as well as handling command line arguments.
	"""
	def __init__(self, stdin="/dev/null", stdout="/dev/null", stderr="/dev/null", pidfile=None, user=None, group=None):
		"""
		The :obj:`stdin`, :obj:`stdout`, and :obj:`stderr` arguments are file
		names that will be opened and be used to replace the standard file
		descriptors in ``sys.stdin``, ``sys.stdout``, and ``sys.stderr``.
		These arguments are optional and default to ``"/dev/null"``. Note that
		stderr is opened unbuffered, so if it shares a file with stdout then
		interleaved output may not appear in the order that you expect.

		:obj:`pidfile` must be the name of a file. :meth:`start` will write
		the pid of the newly forked daemon to this file. :meth:`stop` uses this
		file to kill the daemon.

		:obj:`user` can be the name or uid of a user. :meth:`start` will switch
		to this user for running the service. If :obj:`user` is ``None`` no
		user switching will be done.

		In the same way :obj:`group` can be the name or gid of a group.
		:meth:`start` will switch to this group.
		"""
		self.stdin = stdin
		self.stdout = stdout
		self.stderr = stderr
		self.pidfile = pidfile
		self.user = user
		self.group = group

	def openstreams(self):
		"""
		Open the standard file descriptors stdin, stdout and stderr as specified
		in the constructor.
		"""
		si = open(self.stdin, "rb")
		so = open(self.stdout, "ab+")
		se = open(self.stderr, "ab+", 0)
		os.dup2(si.fileno(), sys.stdin.fileno())
		os.dup2(so.fileno(), sys.stdout.fileno())
		os.dup2(se.fileno(), sys.stderr.fileno())

	def handlesighup(self, signum, frame):
		"""
		Handle a ``SIG_HUP`` signal: Reopen standard file descriptors.
		"""
		self.openstreams()

	def handlesigterm(self, signum, frame):
		"""
		Handle a ``SIG_TERM`` signal: Remove the pid file and exit.
		"""
		if self.pidfile is not None:
			try:
				os.remove(self.pidfile)
			except Exception:
				pass
		sys.exit(0)

	def switchuser(self, user, group):
		"""
		Switch the effective user and group. If :obj:`user` and :obj:`group` are
		both :const:`None` nothing will be done. :obj:`user` and :obj:`group`
		can be an :class:`int` (i.e. a user/group id) or :class:`str`
		(a user/group name).
		"""
		if group is not None:
			if isinstance(group, str):
				group = grp.getgrnam(group).gr_gid
			os.setegid(group)
		if user is not None:
			if isinstance(user, str):
				user = pwd.getpwnam(user).pw_uid
			os.seteuid(user)
			if "HOME" in os.environ:
				os.environ["HOME"] = pwd.getpwuid(user).pw_dir

	def start(self):
		"""
		Daemonize the running script. When this method returns the process is
		completely decoupled from the parent environment.
		"""
		# Finish up with the current stdout/stderr
		sys.stdout.flush()
		sys.stderr.flush()

		# Do first fork
		try:
			pid = os.fork()
			if pid > 0:
				sys.exit(0) # Exit first parent
		except OSError as exc:
			sys.exit("{}: fork #1 failed: ({}) {}\n".format(sys.argv[0], exc.errno, exc.strerror))

		# Decouple from parent environment
		os.chdir("/")
		os.umask(0)
		os.setsid()

		# Do second fork
		try:
			pid = os.fork()
			if pid > 0:
				sys.exit(0) # Exit second parent
		except OSError as exc:
			sys.exit("{}: fork #2 failed: ({}) {}\n".format(sys.argv[0], exc.errno, exc.strerror))

		# Now I am a daemon!

		# Switch user
		self.switchuser(self.user, self.group)

		# Redirect standard file descriptors (will belong to the new user)
		self.openstreams()

		# Write pid file (will belong to the new user)
		if self.pidfile is not None:
			with open(self.pidfile, "wb") as f:
				f.write(str(os.getpid()).encode("utf-8"))

		# Reopen file descriptors on SIGHUP
		signal.signal(signal.SIGHUP, self.handlesighup)

		# Remove pid file and exit on SIGTERM
		signal.signal(signal.SIGTERM, self.handlesigterm)

	def stop(self):
		"""
		Send a ``SIGTERM`` signal to a running daemon. The pid of the daemon
		will be read from the pidfile specified in the constructor.
		"""
		if self.pidfile is None:
			sys.exit("no pidfile specified")
		try:
			with open(self.pidfile, "rb") as f:
				data = f.read()
				pid = int(data.decode("utf-8"))
		except IOError as exc:
			sys.exit("can't open pidfile {}: {}".format(self.pidfile, str(exc)))
		except ValueError:
			sys.exit("mangled pidfile {}: {}".format(self.pidfile, data))
		os.kill(pid, signal.SIGTERM)

	def argparser(self):
		"""
		Return an :mod:`argparse` parser for parsing the command line arguments.
		This can be overwritten in subclasses to add more arguments.
		"""
		p = argparse.ArgumentParser(description="Start, stop or restart a daemon process")
		p.add_argument("action", help="Action to execute", choices=("start", "stop", "restart", "run"))
		p.add_argument("--pidfile", dest="pidfile", help="PID filename (default %(default)s)", default=self.pidfile)
		p.add_argument("--stdin", dest="stdin", help="stdin filename (default %(default)s)", default=self.stdin)
		p.add_argument("--stdout", dest="stdout", help="stdout filename (default %(default)s)", default=self.stdout)
		p.add_argument("--stderr", dest="stderr", help="stderr filename (default %(default)s)", default=self.stderr)
		p.add_argument("--user", dest="user", help="user name or id (default %(default)s)", default=self.user)
		p.add_argument("--group", dest="group", help="group name or id (default %(default)s)", default=self.group)
		return p

	def parseargs(self, parser, args=None):
		"""
		Use the parser returned by :meth:`argparser` to parse the argument
		sequence :obj:`args`, modify :obj:`self` accordingly and return
		the result of the parsers :meth:`parse_args` call.
		"""
		args = parser.parse_args(args)
		self.stdin = args.stdin
		self.stdout = args.stdout
		self.stderr = args.stderr
		self.pidfile = args.pidfile
		self.user = args.user
		self.group = args.group
		return args

	def service(self, args=None):
		"""
		Handle command line arguments and start or stop the daemon accordingly.

		:obj:`args` must be a list of command line arguments (including the
		program name in ``args[0]``). If :obj:`args` is :const:`None` or
		unspecified ``sys.argv`` is used.

		The return value is true when a starting option has been specified as the
		command line argument, i.e. if the daemon should be started.

		The :mod:`argparse` arguments are available afterwards as ``self.args``.
		"""
		p = self.argparser()
		args = self.parseargs(p, args)
		if args.action == "run":
			return True
		elif args.action == "restart":
			self.stop()
			self.start()
			return True
		elif args.action == "start":
			self.start()
			return True
		elif args.action == "stop":
			self.stop()
			return False



#! /usr/bin/env/python
# -*- coding: utf-8 -*-
# cython: language_level=3, always_allow_keywords=True

## Copyright 1999-2016 by LivingLogic AG, Bayreuth/Germany
## Copyright 1999-2016 by Walter Dörwald
##
## All Rights Reserved
##
## See ll/xist/__init__.py for the license


import warnings

from ll.xist import xsc, sims
from ll.xist.ns import html, php


with xsc.Pool():
	class el1(xsc.Element):
		model = sims.Empty()


# The following tests are split into separate test functions, because ``pytest`` has problems otherwise
def test_empty1():
	e = el1()
	e.bytes(validate=True)


def test_empty2():
	e = el1("gurk")
	with warnings.catch_warnings(record=True) as w:
		e.bytes(validate=True)
	assert len(w) == 1
	assert issubclass(w[-1].category, sims.EmptyElementWithContentWarning)


def test_empty3():
	e = el1(php.php("gurk"))
	with warnings.catch_warnings(record=True) as w:
		e.bytes(validate=True)
	assert len(w) == 1
	assert issubclass(w[-1].category, sims.EmptyElementWithContentWarning)


def test_empty4():
	e = el1(xsc.Comment("gurk"))
	with warnings.catch_warnings(record=True) as w:
		e.bytes(validate=True)
	assert len(w) == 1
	assert issubclass(w[-1].category, sims.EmptyElementWithContentWarning)


def test_empty5():
	e = el1(el1())
	with warnings.catch_warnings(record=True) as w:
		e.bytes(validate=True)
	assert len(w) == 1
	assert issubclass(w[-1].category, sims.EmptyElementWithContentWarning)


def test_elements():
	with xsc.Pool():
		class el11(xsc.Element):
			xmlname = "el1"
			xmlns = "ns1"
		class el12(xsc.Element):
			xmlname = "el2"
			xmlns = "ns1"
		class el21(xsc.Element):
			xmlname = "el1"
			xmlns = "ns2"
		class el22(xsc.Element):
			xmlname = "el2"
			xmlns = "ns2"

		el11.model = sims.Elements(el11, el21)

		e = el11()
		e.bytes(validate=True)

		e = el11("foo")
		with warnings.catch_warnings(record=True) as w:
			e.bytes(validate=True)
		assert len(w) == 1
		assert issubclass(w[-1].category, sims.IllegalTextWarning)

		e = el11(php.php("gurk"))
		e.bytes(validate=True)

		e = el11(xsc.Comment("gurk"))
		e.bytes(validate=True)

		e = el11(el11())
		e.bytes(validate=True)

		e = el11(el21())
		e.bytes(validate=True)

		e = el11(el12())
		with warnings.catch_warnings(record=True) as w:
			e.bytes(validate=True)
		assert len(w) == 1
		assert issubclass(w[-1].category, sims.WrongElementWarning)

		e = el11(el22())
		with warnings.catch_warnings(record=True) as w:
			e.bytes(validate=True)
		assert len(w) == 1
		assert issubclass(w[-1].category, sims.WrongElementWarning)


def test_elementsortext():
	with xsc.Pool():
		class el11(xsc.Element):
			xmlname = "el1"
			xmlns = "ns1"
		class el12(xsc.Element):
			xmlname = "el2"
			xmlns = "ns1"
		class el21(xsc.Element):
			xmlname = "el1"
			xmlns = "ns2"
		class el22(xsc.Element):
			xmlname = "el2"
			xmlns = "ns2"

		el11.model = sims.ElementsOrText(el11, el21)

		e = el11()
		e.bytes(validate=True)

		e = el11("foo")
		e.bytes(validate=True)

		e = el11(php.php("gurk"))
		e.bytes(validate=True)

		e = el11(xsc.Comment("gurk"))
		e.bytes(validate=True)

		e = el11(el11())
		e.bytes(validate=True)

		e = el11(el21())
		e.bytes(validate=True)

		e = el11(el12())
		with warnings.catch_warnings(record=True) as w:
			e.bytes(validate=True)
		assert len(w) == 1
		assert issubclass(w[-1].category, sims.WrongElementWarning)

		e = el11(el22())
		with warnings.catch_warnings(record=True) as w:
			e.bytes(validate=True)
		assert len(w) == 1
		assert issubclass(w[-1].category, sims.WrongElementWarning)


def test_noelements():
	with xsc.Pool():
		class el1(xsc.Element):
			xmlns = "ns1"
			model = sims.NoElements()
		class el2(xsc.Element):
			xmlns = "ns2"

		e = el1()
		e.bytes(validate=True)

		e = el1("foo")
		e.bytes(validate=True)

		e = el1(php.php("gurk"))
		e.bytes(validate=True)

		e = el1(xsc.Comment("gurk"))
		e.bytes(validate=True)

		e = el1(el1())
		with warnings.catch_warnings(record=True) as w:
			e.bytes(validate=True)
		assert len(w) == 1
		assert issubclass(w[-1].category, sims.ElementWarning)

		# Elements from a different namespace are OK
		e = el1(el2())
		e.bytes(validate=True)


def test_noelementsortext():
	with xsc.Pool():
		class el1(xsc.Element):
			xmlns = "ns1"
			model = sims.NoElementsOrText()
		class el2(xsc.Element):
			xmlns = "ns2"

		e = el1()
		e.bytes(validate=True)

		e = el1("foo")
		with warnings.catch_warnings(record=True) as w:
			e.bytes(validate=True)
		assert len(w) == 1
		assert issubclass(w[-1].category, sims.IllegalTextWarning)

		e = el1(php.php("gurk"))
		e.bytes(validate=True)

		e = el1(xsc.Comment("gurk"))
		e.bytes(validate=True)

		e = el1(el1())
		with warnings.catch_warnings(record=True) as w:
			e.bytes(validate=True)
		assert len(w) == 1
		assert issubclass(w[-1].category, sims.ElementWarning)

		# Elements from a different namespace are OK
		e = el1(el2())
		e.bytes(validate=True)



#! /usr/bin/env/python
# -*- coding: utf-8 -*-
# cython: language_level=3, always_allow_keywords=True

## Copyright 1999-2016 by LivingLogic AG, Bayreuth/Germany
## Copyright 1999-2016 by Walter Dörwald
##
## All Rights Reserved
##
## See ll/xist/__init__.py for the license


from ll.xist import xsc
from ll.xist.ns import html, xml, chars, abbr, ihtml, wml, specials, htmlspecials, form, meta, svg, fo, docbook, jsp, struts_html, struts_config, tld


def test_variousnamespaces():
	def check(ns, *skip):
		for obj in vars(ns).values():
			if isinstance(obj, type) and issubclass(obj, xsc.Element) and not issubclass(obj, skip):
				node = obj()
				for attrclass in node.Attrs.declaredattrs():
					if attrclass.required:
						if attrclass.values:
							node[attrclass] = attrclass.values[0]
						else:
							node[attrclass] = "foo"
				node.conv().bytes(prefixdefault=True)
		for obj in vars(ns).values():
			if isinstance(obj, type) and issubclass(obj, xsc.Entity) and not issubclass(obj, skip):
				node = obj()
				node.conv().bytes(prefixdefault=True)
		for obj in vars(ns).values():
			if isinstance(obj, type) and issubclass(obj, xsc.ProcInst) and not issubclass(obj, skip):
				node = obj()
				node.conv().bytes(prefixdefault=True)

	check(html)
	check(ihtml)
	check(wml)
	check(specials, specials.include, specials.filetime, specials.filesize)
	check(form)
	check(meta)
	check(htmlspecials, htmlspecials.autoimg, htmlspecials.autopixel)
	check(svg)
	check(fo)
	check(docbook)
	check(jsp)
	check(struts_html, struts_html.taglib) # taglib requires a custom prefix mapping for publishing
	check(struts_config)
	check(tld)


def test_attributeexamples():
	assert xsc.amp.__name__ == "amp"
	assert xsc.amp.xmlname == "amp"
	assert xsc.amp.xmlns is None

	assert chars.uuml.__name__ == "uuml"
	assert chars.uuml.xmlname == "uuml"
	assert chars.uuml.xmlns is None

	assert html.a.Attrs.class_.__name__ == "class_"
	assert html.a.Attrs.class_.xmlname == "class"
	assert html.a.Attrs.class_.xmlns is None

	assert xml.Attrs.lang.__name__ == "lang"
	assert xml.Attrs.lang.xmlname == "lang"
	assert xml.Attrs.lang.xmlns == xml.xmlns



"""ANTLR3 exception hierarchy"""

# begin[licence]
#
# [The "BSD licence"]
# Copyright (c) 2005-2008 Terence Parr
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products
#    derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# end[licence]

from antlr3.constants import INVALID_TOKEN_TYPE


class BacktrackingFailed(Exception):
    """@brief Raised to signal failed backtrack attempt"""

    pass


class RecognitionException(Exception):
    """@brief The root of the ANTLR exception hierarchy.

    To avoid English-only error messages and to generally make things
    as flexible as possible, these exceptions are not created with strings,
    but rather the information necessary to generate an error.  Then
    the various reporting methods in Parser and Lexer can be overridden
    to generate a localized error message.  For example, MismatchedToken
    exceptions are built with the expected token type.
    So, don't expect getMessage() to return anything.

    Note that as of Java 1.4, you can access the stack trace, which means
    that you can compute the complete trace of rules from the start symbol.
    This gives you considerable context information with which to generate
    useful error messages.

    ANTLR generates code that throws exceptions upon recognition error and
    also generates code to catch these exceptions in each rule.  If you
    want to quit upon first error, you can turn off the automatic error
    handling mechanism using rulecatch action, but you still need to
    override methods mismatch and recoverFromMismatchSet.
    
    In general, the recognition exceptions can track where in a grammar a
    problem occurred and/or what was the expected input.  While the parser
    knows its state (such as current input symbol and line info) that
    state can change before the exception is reported so current token index
    is computed and stored at exception time.  From this info, you can
    perhaps print an entire line of input not just a single token, for example.
    Better to just say the recognizer had a problem and then let the parser
    figure out a fancy report.
    
    """

    def __init__(self, input=None):
        Exception.__init__(self)

	# What input stream did the error occur in?
        self.input = None

        # What is index of token/char were we looking at when the error
        # occurred?
        self.index = None

	# The current Token when an error occurred.  Since not all streams
	# can retrieve the ith Token, we have to track the Token object.
	# For parsers.  Even when it's a tree parser, token might be set.
        self.token = None

	# If this is a tree parser exception, node is set to the node with
	# the problem.
        self.node = None

	# The current char when an error occurred. For lexers.
        self.c = None

	# Track the line at which the error occurred in case this is
	# generated from a lexer.  We need to track this since the
        # unexpected char doesn't carry the line info.
        self.line = None

        self.charPositionInLine = None

        # If you are parsing a tree node stream, you will encounter som
        # imaginary nodes w/o line/col info.  We now search backwards looking
        # for most recent token with line/col info, but notify getErrorHeader()
        # that info is approximate.
        self.approximateLineInfo = False

        
        if input is not None:
            self.input = input
            self.index = input.index()

            # late import to avoid cyclic dependencies
            from antlr3.streams import TokenStream, CharStream
            from antlr3.tree import TreeNodeStream

            if isinstance(self.input, TokenStream):
                self.token = self.input.LT(1)
                self.line = self.token.line
                self.charPositionInLine = self.token.charPositionInLine

            if isinstance(self.input, TreeNodeStream):
                self.extractInformationFromTreeNodeStream(self.input)

            else:
                if isinstance(self.input, CharStream):
                    self.c = self.input.LT(1)
                    self.line = self.input.line
                    self.charPositionInLine = self.input.charPositionInLine

                else:
                    self.c = self.input.LA(1)

    def extractInformationFromTreeNodeStream(self, nodes):
        from antlr3.tree import Tree, CommonTree
        from antlr3.tokens import CommonToken
        
        self.node = nodes.LT(1)
        adaptor = nodes.adaptor
        payload = adaptor.getToken(self.node)
        if payload is not None:
            self.token = payload
            if payload.line <= 0:
                # imaginary node; no line/pos info; scan backwards
                i = -1
                priorNode = nodes.LT(i)
                while priorNode is not None:
                    priorPayload = adaptor.getToken(priorNode)
                    if priorPayload is not None and priorPayload.line > 0:
                        # we found the most recent real line / pos info
                        self.line = priorPayload.line
                        self.charPositionInLine = priorPayload.charPositionInLine
                        self.approximateLineInfo = True
                        break
                    
                    i -= 1
                    priorNode = nodes.LT(i)
                    
            else: # node created from real token
                self.line = payload.line
                self.charPositionInLine = payload.charPositionInLine
                
        elif isinstance(self.node, Tree):
            self.line = self.node.line
            self.charPositionInLine = self.node.charPositionInLine
            if isinstance(self.node, CommonTree):
                self.token = self.node.token

        else:
            type = adaptor.getType(self.node)
            text = adaptor.getText(self.node)
            self.token = CommonToken(type=type, text=text)

     
    def getUnexpectedType(self):
        """Return the token type or char of the unexpected input element"""

        from antlr3.streams import TokenStream
        from antlr3.tree import TreeNodeStream

        if isinstance(self.input, TokenStream):
            return self.token.type

        elif isinstance(self.input, TreeNodeStream):
            adaptor = self.input.treeAdaptor
            return adaptor.getType(self.node)

        else:
            return self.c

    unexpectedType = property(getUnexpectedType)
    

class MismatchedTokenException(RecognitionException):
    """@brief A mismatched char or Token or tree node."""
    
    def __init__(self, expecting, input):
        RecognitionException.__init__(self, input)
        self.expecting = expecting
        

    def __str__(self):
        #return "MismatchedTokenException("+self.expecting+")"
        return "MismatchedTokenException(%r!=%r)" % (
            self.getUnexpectedType(), self.expecting
            )
    __repr__ = __str__


class UnwantedTokenException(MismatchedTokenException):
    """An extra token while parsing a TokenStream"""

    def getUnexpectedToken(self):
        return self.token


    def __str__(self):
        exp = ", expected %s" % self.expecting
        if self.expecting == INVALID_TOKEN_TYPE:
            exp = ""

        if self.token is None:
            return "UnwantedTokenException(found=%s%s)" % (None, exp)

        return "UnwantedTokenException(found=%s%s)" % (self.token.text, exp)
    __repr__ = __str__


class MissingTokenException(MismatchedTokenException):
    """
    We were expecting a token but it's not found.  The current token
    is actually what we wanted next.
    """

    def __init__(self, expecting, input, inserted):
        MismatchedTokenException.__init__(self, expecting, input)

        self.inserted = inserted


    def getMissingType(self):
        return self.expecting


    def __str__(self):
        if self.inserted is not None and self.token is not None:
            return "MissingTokenException(inserted %r at %r)" % (
                self.inserted, self.token.text)

        if self.token is not None:
            return "MissingTokenException(at %r)" % self.token.text

        return "MissingTokenException"
    __repr__ = __str__


class MismatchedRangeException(RecognitionException):
    """@brief The next token does not match a range of expected types."""

    def __init__(self, a, b, input):
        RecognitionException.__init__(self, input)

        self.a = a
        self.b = b
        

    def __str__(self):
        return "MismatchedRangeException(%r not in [%r..%r])" % (
            self.getUnexpectedType(), self.a, self.b
            )
    __repr__ = __str__
    

class MismatchedSetException(RecognitionException):
    """@brief The next token does not match a set of expected types."""

    def __init__(self, expecting, input):
        RecognitionException.__init__(self, input)

        self.expecting = expecting
        

    def __str__(self):
        return "MismatchedSetException(%r not in %r)" % (
            self.getUnexpectedType(), self.expecting
            )
    __repr__ = __str__


class MismatchedNotSetException(MismatchedSetException):
    """@brief Used for remote debugger deserialization"""
    
    def __str__(self):
        return "MismatchedNotSetException(%r!=%r)" % (
            self.getUnexpectedType(), self.expecting
            )
    __repr__ = __str__


class NoViableAltException(RecognitionException):
    """@brief Unable to decide which alternative to choose."""

    def __init__(
        self, grammarDecisionDescription, decisionNumber, stateNumber, input
        ):
        RecognitionException.__init__(self, input)

        self.grammarDecisionDescription = grammarDecisionDescription
        self.decisionNumber = decisionNumber
        self.stateNumber = stateNumber


    def __str__(self):
        return "NoViableAltException(%r!=[%r])" % (
            self.unexpectedType, self.grammarDecisionDescription
            )
    __repr__ = __str__
    

class EarlyExitException(RecognitionException):
    """@brief The recognizer did not match anything for a (..)+ loop."""

    def __init__(self, decisionNumber, input):
        RecognitionException.__init__(self, input)

        self.decisionNumber = decisionNumber


class FailedPredicateException(RecognitionException):
    """@brief A semantic predicate failed during validation.

    Validation of predicates
    occurs when normally parsing the alternative just like matching a token.
    Disambiguating predicate evaluation occurs when we hoist a predicate into
    a prediction decision.
    """

    def __init__(self, input, ruleName, predicateText):
        RecognitionException.__init__(self, input)
        
        self.ruleName = ruleName
        self.predicateText = predicateText


    def __str__(self):
        return "FailedPredicateException("+self.ruleName+",{"+self.predicateText+"}?)"
    __repr__ = __str__
    

class MismatchedTreeNodeException(RecognitionException):
    """@brief The next tree mode does not match the expected type."""

    def __init__(self, expecting, input):
        RecognitionException.__init__(self, input)
        
        self.expecting = expecting

    def __str__(self):
        return "MismatchedTreeNodeException(%r!=%r)" % (
            self.getUnexpectedType(), self.expecting
            )
    __repr__ = __str__
